#include "config.h"
#include "request.h"
#include "com.h"
#include "crypto.h"
#include "mess.h"
#include "iv.h"

#include "command.h"
#include "list.h"

#include "statistics.h"

#include "sha1.h"


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <zlib.h>

static buf cryptbuf;
static buf _command;
static buf *command = &_command;

static buf compress_buf;
static void compress_buffer(buf *b)
{
  char cb[512];
  buf t;
  z_stream z;
  int i;

  buf_reset(&compress_buf);

  z.zalloc = Z_NULL;
  z.zfree = Z_NULL;
  z.opaque = Z_NULL;

  if (deflateInit(&z, 9) != Z_OK) goto fatal_zlib;

  z.next_in = (unsigned char *)b->buf;
  z.avail_in = b->size;
  z.next_out = (unsigned char *)cb;
  z.avail_out = 512;

#define EMPTY_OUT				\
  do {						\
    for (i=0; i < 512 - z.avail_out; i++)	\
      buf_add_char(&compress_buf, cb[i]);	\
    z.next_out = (unsigned char *)cb;		\
    z.avail_out = 512;				\
  } while (0)

  while (1) {
    int r = deflate(&z, 0);
    if (r == Z_STREAM_END) break;
    if (r != Z_OK) goto fatal_zlib;
    EMPTY_OUT;
    if (z.avail_in == 0) break;
  }
  EMPTY_OUT;
  while (1) {
    int r = deflate(&z, Z_FINISH);
    if (r == Z_STREAM_END) break;
    if (r != Z_OK) goto fatal_zlib;
    EMPTY_OUT;
  }
  EMPTY_OUT;

  if (deflateEnd(&z) != Z_OK) goto fatal_zlib;

  /* exchange b and compress_buf */
  t = *b;
  *b = compress_buf;
  compress_buf = t;

  return;

fatal_zlib:
  printf("zlib fatal error\n");
  exit(1);
#undef EMPTY_OUT
}

/* the buffer may contain at most 64 bytes of random at first, so
 * we try again and again
 */
static void uncompress_buffer(buf *b)
{
  char cb[512];
  buf t;
  z_stream z;
  int i;
  int pos = 0;
  char *da_buf = b->buf;

try_again:
  buf_reset(&compress_buf);

  z.zalloc = Z_NULL;
  z.zfree = Z_NULL;
  z.opaque = Z_NULL;

  if (inflateInit(&z) != Z_OK) {
    mess("ZLIB inflateInit error\n");
    return;
  }

  z.next_in = (unsigned char *)da_buf;
  z.avail_in = b->size - pos;
  z.next_out = (unsigned char *)cb;
  z.avail_out = 512;

#define EMPTY_OUT				\
  do {						\
    for (i=0; i < 512 - z.avail_out; i++)	\
      buf_add_char(&compress_buf, cb[i]);	\
    z.next_out = (unsigned char *)cb;		\
    z.avail_out = 512;				\
  } while (0)

  while (1) {
    int r = inflate(&z, Z_SYNC_FLUSH);
    if (r == Z_STREAM_END) break;
    if (r == Z_BUF_ERROR) break;
    if (r != Z_OK) {
      inflateEnd(&z);
      da_buf++;
      pos++;
      if (pos == 65) {
        mess("ZLIB inflate error %d\n", r);
        return;
      }
      goto try_again;
    }
    EMPTY_OUT;
  }
  EMPTY_OUT;

  if (inflateEnd(&z) != Z_OK) {
    da_buf++;
    pos++;
    if (pos == 65) {
      mess("ZLIB inflateEnd error\n");
      return;
    }
    goto try_again;
  }

  mess("uncompressed at turn %d\n", pos);

  /* exchange b and compress_buf */
  t = *b;
  *b = compress_buf;
  compress_buf = t;
#undef EMPTY_OUT
}

static void normalize(buf *b)
{
  unsigned char sha[20];
  SHA1Context ctx;
  int length;
  int len;
  int pos;

  /* add 20 bytes of random */
  for (len = 0; len < 20; len++) buf_add_char(b, rand()&0xff);

  /* add sha1(K, sha1(K, M)) */
  /* first sha1(K, M) */
  SHA1Reset(&ctx);
  SHA1Input(&ctx, get_key(), 32);
  SHA1Input(&ctx, (unsigned char *)b->buf, b->size);
  SHA1Result(&ctx, sha);
  /* then sha1(K, sha1(K, M)) */
  SHA1Reset(&ctx);
  SHA1Input(&ctx, get_key(), 32);
  SHA1Input(&ctx, sha, 20);
  SHA1Result(&ctx, sha);
  for (len = 0; len < 20; len++) buf_add_char(b, sha[len]);

  request_data_size = b->size;

  /* compress */
  compress_buffer(b);

  /* add random */
  length = b->size + 8;
  len = 64 - length % 64;
  if (len == 64) len = 0;
  len += 8;
  while (len) { buf_add_char(b, rand() & 0xff); len--; }

  /* displace data by putting random at start or end */
  /* we put 64 bytes, to keep size%64==0 */
  buf_ensure_size(b, b->size+64);
  /* maybe not very good this % */
  pos = rand() % 65;
  /* brutal memmove, I should do it more cleverly */
  memmove(b->buf + pos, b->buf, b->size);
  for (len = 0; len < pos; len++) b->buf[len] = rand() & 0xff;
  for (len = 0; len < 64-pos; len++) b->buf[len+pos+b->size] = rand() & 0xff;
  b->size += 64;

  mess("message displaced at position %d\n", pos);

  request_data_zip_size = b->size;
}

static void verify_input(buf *b)
{
  unsigned char sha[20];
  SHA1Context ctx;
  int i;

  if (b->size <= 40) goto verif_error;

  SHA1Reset(&ctx);
  SHA1Input(&ctx, get_key(), 32);
  SHA1Input(&ctx, (unsigned char *)b->buf, b->size-20);
  SHA1Result(&ctx, sha);
  SHA1Reset(&ctx);
  SHA1Input(&ctx, get_key(), 32);
  SHA1Input(&ctx, sha, 20);
  SHA1Result(&ctx, sha);

  mess("remsha: "); fmess(b->buf + b->size-20, 20); mess("\n");
  mess("locsha: "); fmess(sha, 20); mess("\n");

  for (i = 0; i < 20; i++)
    if ((unsigned char)b->buf[b->size-20+i] != sha[i]) goto verif_error;

  return;

verif_error:
  mess("Bad reply from server:\n");
  mess("buf size = %d\n", b->size);
  mess("BEGIN mess:\n");
  fmess(b->buf, b->size);
  mess("\nEND mess.\n");
#if 0
  /* out IV is fucked up, get the correct one back */
  iv(b);
  if (b->size != 16*2) {
    mess("bad IV from server:\n");
    if (!b->size)
      mess("(null)\n");
    else
      fmess(b->buf, b->size);
  } else
    init_crypto(b->buf);
#endif
  buf_reset(b);
}

/* this variable is modified by the function 'synchronize_iv' in iv.c
 * and only by the function 'request' in this file.
 */
int must_synchronize_iv;
/* do_cypher = 0 if no encryption performed
 *           = 1 if both request and reply are encrypted
 *           = 2 if only reply is encrypted
 */
static void request(buf *in, buf *out, int do_cypher)
{
  if (must_synchronize_iv) {
    mess("A request has failed. You must synchronize with the server.\n");
    buf_reset(out);
    return;
  }

  r_reset();
  b_reset();

  r_add_string("POST ");
  r_add_string(server_url);
  r_add_string(" HTTP/1.1");
  r_add_eol();

  r_add_string("Host: ");
  r_add_string(server);
  if (port != 80) {
    r_add_string(":");
    r_add_int(port);
  }
  r_add_eol();

  r_add_string("Accept: */*"); r_add_eol();
  r_add_string("Connection: close"); r_add_eol();
  r_add_string("User-Agent: "); r_add_string(VERSION); r_add_eol();
  r_add_string("MIME-Version: 1.0"); r_add_eol();

  r_add_string("Content-Type: multipart/form-data; boundary=");
  r_add_string(BOUNDARY);
  r_add_eol();

  b_add_string("--"); b_add_string(BOUNDARY); b_add_eol();

  b_add_string(
    "Content-Disposition: form-data; name=\"files\"; filename=\"data.bin\"");
  b_add_eol();
  b_add_string("Content-Type: application/octet-stream"); b_add_eol();
  b_add_string("Content-Transfer-Encoding: binary"); b_add_eol();
  b_add_eol();

  /* add data */
  if (do_cypher == 1) {
    normalize(in);
    buf_ensure_size(&cryptbuf, in->size);
    cypher(cryptbuf.buf, in->buf, in->size);
    save_iv();
    b_add_buffer(cryptbuf.buf, in->size);
  } else
    b_add_buffer(in->buf, in->size);

  b_add_eol();

  b_add_string("--"); b_add_string(BOUNDARY); b_add_string("--");
  b_add_eol();

  /* finish request */
  r_add_string("Content-Length: ");
  r_add_int(b_size());
  r_add_eol();
  r_add_eol();

  r_add_buffer(b_req(), b_size());

  if (do_cypher != 0) {
    post(&cryptbuf);
    buf_ensure_size(out, cryptbuf.size);
    cypher(out->buf, cryptbuf.buf, cryptbuf.size);
    out->size = cryptbuf.size;

    request_reply_data_zip_size = out->size;

    uncompress_buffer(out);
    verify_input(out);

    request_reply_data_size = out->size;
  } else {
    post(out);
    /* non-cyphered request is IV, which has an HMAC too */
    verify_input(out);
    request_reply_data_size = out->size;
    request_reply_data_zip_size = 0;
  }

  if (out->size == 0) {
    must_synchronize_iv = 1;
    mess("Request failed. You must synchronize with the server.\n");
  } else
    save_iv();

#if 0
  mess("Message sent:\n");
  fmess(r_req(), r_size());
  mess("End of sent message\n");

  mess("reply size: %d\n", out->size);
#endif
}

void iv(buf *out)
{
  int i;
  buf in;
  in.size = 2;
  in.maxsize = 2;
  in.buf = "IV";
  request(&in, out, 0);
  /* check that IV is ok */
  /* good size */
  if (out->size != 32+20) {
iv_err:
    buf_reset(out);
    return;
  }
  /* good data */
  for (i = 0; i < 32; i++) {
    char t = out->buf[i];
    if (!((t >= '0' && t <= '9') ||
          (t >= 'a' && t <= 'f') ||
          (t >= 'A' && t <= 'F')))
      goto iv_err;
  }
}

/* return out->size = 1 if ok, out->size = 0 if error */
/* the crypto-engine must be initialized with the IV before calling iw */
void iw(buf *out, char *iv)
{
  char _in[2+16+20];
  buf in;
  SHA1Context ctx;

  in.size = 2+16+20;
  in.maxsize = 2+16+20;
  in.buf = _in;

  _in[0] = 'I';
  _in[1] = 'W';

  memcpy(_in+2, iv, 16);

  SHA1Reset(&ctx);
  SHA1Input(&ctx, get_key(), 32);
  SHA1Input(&ctx, (unsigned char *)_in, 16+2);
  SHA1Result(&ctx, (unsigned char *)_in+2+16);
  SHA1Reset(&ctx);
  SHA1Input(&ctx, get_key(), 32);
  SHA1Input(&ctx, (unsigned char *)_in+2+16, 20);
  SHA1Result(&ctx, (unsigned char *)_in+2+16);

  request(&in, out, 2);

  if (out->size == 0 ||
      out->buf[0] != 'O' ||
      out->buf[1] != 'K') goto error;

  out->size = 1;
  return;

error:
  out->size = 0;
}

/* return q to NULL on error, passed >=0 int if ok */
static int get_int(char **q)
{
  /* lovely pointers, no? enjoy */
  int l;
  if (!isdigit(**q)) return 0;
  if (sscanf(*q, "%d", &l) != 1) goto error;
  if (l < 0) goto error;
  while (isdigit(**q)) (*q)++;
  if (**q != ' ') goto error;
  (*q)++;
  return l;
error:
  *q = 0;
  return 0;
}

/* set q to NULL on error, passed string if ok */
static char *get_string(char **q)
{
  int i;
  char *ret;
  int size = get_int(q); if (!*q) goto error;
  if (size <= 0) goto error;

  /* no 0 (meaning end of q) in the string allowed */
  for (i = 0; i < size; i++) if ((*q)[i] == 0) goto error;

  ret = malloc(size+1);
  if (!ret) goto error;
  memcpy(ret, *q, size);
  ret[size] = 0;
  *q += size;
  return ret;

error:
  *q = 0;
  return 0;
}

void ls(buf *out, char *dir, list *ls_list)
{
  int nb_files;
  int i;
  char *z;
  char *q;
  c_reset();
  c_add_string("LS");
  c_add_int(strlen(dir));
  c_add_string(" ");
  c_add_string(dir);
  request(command, out, 1);

  //mess("reply: %s\n", z);
  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;
  q = z+2;

  /* q points to number of files */
  nb_files = get_int(&q); if (!q) goto error;

  empty_list(ls_list);
  for (i = 0; i < nb_files; i++) {
    int size;
    char *file_name;
    /* get file name */
    file_name = get_string(&q); if (!q) goto error;
    list_add(ls_list, file_name);
    size = get_int(&q); if (!q) { free(file_name); goto error; }
    list_add_hidden_int(ls_list, size);
    mess("%10d %s\n", size, file_name);
    free(file_name);
  }

  return;

error:
  mess("Bad reply from server.\n");
  buf_reset(out);
  empty_list(ls_list);
  list_add(ls_list, "./");
}

/* set out->size = 0 on error, =1 if all ok */
void gt(buf *out, char *local_file, char *file, int start, int length)
{
  char *z;
  char *q;
  FILE *f;
  struct stat b;
  c_reset();
  c_add_string("GT");
  c_add_int(strlen(file));
  c_add_string(" ");
  c_add_string(file);
  c_add_int(start);
  c_add_string(" ");
  c_add_int(length);
  c_add_string(" ");
  request(command, out, 1);

  mess("reply (len=%d):\n", out->size);
  fmess(out->buf, out->size);
  mess("\nend of reply\n");

  if (out->size + 2 < length) goto error;

  //mess("reply:\n");
  //fmess(z, out->size);
  //mess("\nend of reply\n");

  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;
  q = z+2;

  /* out->size-1 because we added a 0 */
  if ((q-z) + length > out->size-1) goto error;

  if (stat(local_file, &b) != -1)
    f = fopen(local_file, "r+b");
  else
    f = fopen(local_file, "w+b");
  if (!f) goto error_local;
  if (fseek(f, start, SEEK_SET) != 0) { fclose(f); goto error_local; }
  if (fwrite(q, length, 1, f) != 1) { fclose(f); goto error_local; }
  if (fclose(f) != 0) goto error_local;

  out->size = 1;
  return;

error:
  mess("Bad reply from server.\n");
  out->size = 0;
  return;

error_local:
  mess("Bad local file: %s\n", local_file);
  out->size = 0;
}

/* set out->size = 0 on error, =1 if all ok */
void pt(buf *out, char *local_file, char *file, int start, int length)
{
  char *z;
  char *buf = malloc(length);
  FILE *f = 0;
  if (!buf) { printf("out of memory\n"); exit(1); }

  c_reset();
  c_add_string("PT");
  c_add_int(strlen(file));
  c_add_string(" ");
  c_add_string(file);
  c_add_int(start);
  c_add_string(" ");
  c_add_int(length);
  c_add_string(" ");

  f = fopen(local_file, "rb");
  if (!f) goto error_local;
  if (fseek(f, start, SEEK_SET) != 0) goto error_local;
  if (fread(buf, length, 1, f) != 1) goto error_local;
  if (fclose(f) != 0) { f = 0; goto error_local; }
  f = 0;
  c_add_buffer(buf, length);

  request(command, out, 1);

  //mess("reply: %s\n", z);

  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;

  free(buf);
  if (f) fclose(f);
  out->size = 1;
  return;

error:
  free(buf);
  mess("Bad reply from server.\n");
  if (f) fclose(f);
  out->size = 0;
  return;

error_local:
  free(buf);
  mess("Bad local file: %s\n", local_file);
  if (f) fclose(f);
  out->size = 0;
}

void md(buf *out, char *dir)
{
  char *z;

  c_reset();
  c_add_string("MD");
  c_add_int(strlen(dir));
  c_add_string(" ");
  c_add_string(dir);

  request(command, out, 1);

  buf_add_char(out, 0);
  z = out->buf;

  //mess("reply: %s\n", z);

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;

  return;

error:
  mess("Bad reply from server.\n");
}

void dd(buf *out, char *dir)
{
  char *z;

  c_reset();
  c_add_string("DD");
  c_add_int(strlen(dir));
  c_add_string(" ");
  c_add_string(dir);

  request(command, out, 1);

  //mess("reply: %s\n", z);
  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;

  return;

error:
  mess("Bad reply from server.\n");
}

void df(buf *out, char *file)
{
  char *z;

  c_reset();
  c_add_string("DF");
  c_add_int(strlen(file));
  c_add_string(" ");
  c_add_string(file);

  request(command, out, 1);

  //mess("reply: %s\n", z);
  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;

  return;

error:
  mess("Bad reply from server.\n");
}

void rn(buf *out, char *oldname, char *newname)
{
  char *z;

  c_reset();
  c_add_string("RN");
  c_add_int(strlen(oldname));
  c_add_string(" ");
  c_add_string(oldname);
  c_add_int(strlen(newname));
  c_add_string(" ");
  c_add_string(newname);

  request(command, out, 1);

  //mess("reply: %s\n", z);
  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;

  return;

error:
  mess("Bad reply from server.\n");
}

/* return -1 on error */
int fs(buf *out,  char *file)
{
  char *z;
  char *q;
  int len;

  c_reset();
  c_add_string("FS");
  c_add_int(strlen(file));
  c_add_string(" ");
  c_add_string(file);

  request(command, out, 1);

  //mess("reply: %s\n", z);
  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;
  q = z+2;
  len = get_int(&q);
  if (!q) goto error;
  if (len < 0) goto error;

  return len;

error:
  mess("Bad reply from server.\n");
  return -1;
}

/* set out->size = 0 on error, = 1 if ok */
void sh(buf *out,  char *file, char *outsha)
{
  char *z;

  c_reset();
  c_add_string("SH");
  c_add_int(strlen(file));
  c_add_string(" ");
  c_add_string(file);

  request(command, out, 1);

  //mess("reply: %s\n", z);
  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;
  if (out->size < 22) goto error;
  memcpy(outsha, z+2, 20);

  out->size = 1;
  return;

error:
  mess("Bad reply from server.\n");
  memset(outsha, 0, 20);
  out->size = 0;
  return;
}

void ml(buf *out)
{
  char *z;

  c_reset();
  c_add_string("ML");

  request(command, out, 1);

  //mess("reply: %s\n", z);
  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;
  mess("%s\n", z+2);

  return;

error:
  mess("Bad reply from server.\n");
  return;
}

void px(buf *out, char *server, int port, buf *in)
{
  int len;
  char *z;
  char *q;
  c_reset();
  c_add_string("PX");
  c_add_int(strlen(server));
  c_add_string(" ");
  c_add_string(server);
  c_add_int(port);
  c_add_string(" ");
  c_add_int(in->size);
  c_add_string(" ");
  c_add_buffer(in->buf, in->size);
  request(command, out, 1);

  //mess("reply: %s\n", z);
  buf_add_char(out, 0);
  z = out->buf;

  if (z[0] == 'E' && z[1] == 'R') {
    mess("**error: %s\n", z+2);
    goto error;
  }

  if (z[0] != 'O' || z[1] != 'K') goto error;
  q = z+2;

  /* q points to size of reply */
  len = get_int(&q); if (!q) goto error;
  /* size-1 because we added a 0 */
  if (len + (q-z) > out->size-1) goto error;

  //buf_reset(out);
  //for (i = 0; i < len; i++) buf_add_char(out, q[i]);
  memmove(out->buf, q, len);
  out->size = len;

  return;

error:
  mess("Bad reply from server.\n");
  buf_reset(out);
}
